
################################################################
# proc readIncludeFile {arrayName fileName}--
#    Read an include file, and make a lookup table for all the
#    lines that match a 
#    #define xxx val  /* comment */
#    or 
#    #define xxx val
#    format
#
# Arguments
#   arrayName: The name of an array in the calling procedures scope.
#   fileName: The full path for the include file to be read.
# 
# Results
#   Adds new indices to the arrayName in the form:
#   set arrayName(mnemonic.value) "comment"
# 
proc readIncludeFile {arrayName fileName} {
  upvar $arrayName localArray

  set if [open $fileName r]

  while {[set len [gets $if line]] >= 0} {
    
    # Use the Advanced Regular Expression "\s"
    # Class Shorthand to encode for whitespace.
    set m [regexp \
        {#define[\s]+([^\s]+)[\s]+([^\s]+)[\s]+/\*([^*]+) \*/} \
        $line a mnemonic num comment]
    
    # If no match, it might be a line with no comment
    if {!$m} {
      set m [regexp {#define\s+([^\s]+)\s+([^\s]+)} \
        $line a mnemonic num]
      set comment "No Comment"
    }

    # If still no match, this line must note be a #define line
    #  skip it.
    if {!$m} {continue}
    
    # Holler if there's a collision.  I guess a human will
    #  need to deal with this file.
    if {[info exists localArray($mnemonic.$num)]} {
       puts "COLLISION: $num"
       exit
    }

    # Everyone's happy, assign the value
    set localArray($mnemonic.$num) $comment
  }
}

if {[info exists argv] && [string match $argv "-test"] } {
  readIncludeFile Lookups /usr/include/arpa/telnet.h

  foreach index [lsort [array names Lookups]] {
    puts "$index $Lookups($index)"
  }
}
