source binaryStrings.tcl
source readincl.tcl

namespace eval telnet {
    variable Telnet
    set debug 0
    readIncludeFile Lookups /usr/include/arpa/telnet.h
    set Lookups(UNKNOWN.-1) "Unknown Option"

    proc Debugputs {string} {
	variable debug
	if {$debug} {
	    puts $string
	}
    }

    proc splitOOB {string} {
	variable Lookups
	variable Telnet

	set text ""
	set oob ""

	# Set empty the commentary.
	set Telnet(comment) ""

	set lst [binaryStrings::bin2hex $string]

	for {set i 0} {$i < [llength $lst]} {incr i} {
	    set l [lindex $lst $i]
	    while {[string match $l "ff"]} {
		set Telnet(comment-temp) ""
		incr i
		set m [lindex $lst $i]
		scan $m %x dec
		incr i
		incr i [$m [lrange $lst $i end] text oob]
		set id [array names Lookups *.$dec]
		Debugputs "NEW M: $m :: $dec :: [array names Lookups *.$dec]\
		  :: $oob"
		lappend Telnet(comment) "$Lookups($id): $Telnet(comment-temp)"
		set l [lindex $lst $i]
		Debugputs "NEW I: $i :: NEW-L $l"
	    }
	    append text [string trim $l]
	}
	return [list [binaryStrings::hex2bin $text]\
	  [binaryStrings::hex2bin $oob]]
    }

    proc initialize {s} {
	variable Telnet
	# No initialization of connection being done.
    }

    proc binarySend {s str} {
	if {[string length $str] == 0} {return}
	set config [fconfigure $s -translation]
	fconfigure $s -translation binary
	puts -nonewline $s $str
	flush $s
	fconfigure $s -translation $config
    }

    #
    # The procedures for dealing with individual IAC options
    #

    proc f0 {subl textName oobName} {
	Debugputs "End of Negotiations - This should not happen"
	return 1
    }

    # Receive a NOOP - ignore it.
    proc f1 {} {uplevel {incr i} }

    # Receive a Data Mark - ignore it.
    proc f2 {} { uplevel {incr i} }

    # Receive a Break - ignore it.
    proc f3 {} { uplevel {incr i} }

    # Receive a Interrupt - ignore it.
    proc f4 {} { uplevel {incr i} }

    # Receive a Abort output - ignore it.
    proc f5 {} { uplevel {incr i} }

    # Are you there - send a NO-OP just to say hello.
    proc f6 {subl textName oobName} {
	upvar $oobName oob
	append oob fff1
    }

    # Receive a Erase character - ignore it.
    proc f7 {} { uplevel {incr i} }

    # Receive a Erase line - ignore it.
    proc f8 {} { uplevel {incr i} }

    # Receive a go ahead - ignore it.
    proc f9 {} { uplevel {incr i} }

    # Do the fa Negotiation
    proc fa {subl text oob} {
	# Starts with character after the 'fa'
	Debugputs "Dealing with: [lrange $subl 0 20]"
	set count 1
	set type [lindex $subl 0]
	switch $type {
	18 {
		foreach {count string} [findFFF0 $subl] {  }
		append Telnet(comment-temp) " Terminal Type: $string"
		return $count
	    }
	1f {
		set wid1 [lindex $subl 1]
		set wid2 [lindex $subl 2]
		set ht1 [lindex $subl 3]
		set ht2 [lindex $subl 4]
		Debugputs "DISPLAY IS [expr 0x${ht1}${ht2}] x\
		  [expr 0x${wid1}${wid2}]"
		append Telnet(comment-temp) " DISPLAY IS [expr 0x${ht1}${ht2}]\
		  x [expr 0x${wid1}${wid2}]"
		# Add 2 more for the ff f0 end negotiation
		return 7
	    }
	22 {
		# Line mode complexity.  Just find the end.
		foreach {count string} [findFFF0 $subl] {  }
		append Telnet(comment-temp) " Terminal Type: $string"
		return $count
	    }
	default {
		Debugputs "NOT DEALING WITH $type yet"
	    }
	}
    }


    # WILL command fb : Reply DO(fd) or DONT(fe)
    proc fb {subl textName oobName} {
	upvar $oobName oob
	variable Telnet
	set opt [lindex $subl 0]
	append oob $Telnet(will.reply.$opt)

	addInfo $subl TELOPT_

	if {[info exists Telnet(next.will.reply.$opt)]} {
	    set Telnet(will.reply.$opt) $Telnet(next.will.reply.$opt)
	}
	return 1
    }

    # WONT(fc) - MUST REPLY DONT (fe)
    proc fc {subl textName oobName} {
	variable Telnet
	upvar $oobName oob
	if {$Telnet(wont.[lindex $subl 0]) < 3} {
	    append oob fffe[lindex $subl 0]
	}
	addInfo $subl TELOPT_
	incr Telnet(wont.[lindex $subl 0])
	return 1
    }

    # DO command fc : REply WILL(fb) or WONT (fc) (or values for 1f)
    array set Telnet { do.reply.00  {fffc00} do.reply.01  {fffc01}
do.reply.03  {fffb03} do.reply.05  {fffc05} do.reply.06  {fffc06}
do.reply.18  {fffa1864756d62fff0} do.reply.1f  {fffa00800080fff0} do.reply.20 \
      {fffc20}
do.reply.21  {fffc21} do.reply.22  {fffc22} do.reply.23  {fffc23}
do.reply.24  {fffc24} do.reply.25  {fffc25} do.reply.26  {fffc26}
do.reply.27  {fffc27} do.reply.28  {fffc28} do.reply.2c  {fffc2c}
}


    # WILL command fb : Reply DO(fd) or DONT(fe)
    array set Telnet { will.reply.01  {fffd01} will.reply.03  {fffd03}
will.reply.05  {fffe05} will.reply.06  {fffe06} will.reply.18  {fffe18}
will.reply.1f  {fffd1f} will.reply.20  {fffe20} will.reply.21  {fffe21}
will.reply.22  {fffe22} will.reply.23  {fffe23} will.reply.24  {fffe24}
will.reply.25  {fffe25} will.reply.26  {fffe26} will.reply.27  {fffe27}
will.reply.28  {fffe28} will.reply.2c  {fffe2c}
}

    array set Telnet {next.will.reply.1f fffa1f00500080ff00}

    # Counts for dont/wont discussion.  One should not get into a
    #   I-already-told-you-that match with the server.

    array set Telnet { dont.00  0 dont.01  0 dont.03  0 dont.05  0 dont.06 
   0 dont.18  0 dont.1f  0 dont.20  0 dont.21  0 dont.22  0 dont.23  0
   dont.24  0 dont.25  0 dont.26  0 dont.27  0 dont.28  0 dont.2c  0
   wont.00  0 wont.01  0 wont.03  0 wont.05  0 wont.06 
   0 wont.18  0 wont.1f  0 wont.20  0 wont.21  0 wont.22  0 wont.23  0
   wont.24  0 wont.25  0 wont.26  0 wont.27  0 wont.28  0 wont.2c  0}

    # DO command fd : Reply WILL(fb) or WONT (fc)
    proc fd {subl textName oobName} {
	variable Telnet
	upvar $oobName oob

	addInfo $subl TELOPT_

	append oob $Telnet(do.reply.[lindex $subl 0])

	return 1
    }

    proc addInfo {subl prefix} {
	variable Telnet
	variable Lookups

	set m [lindex $subl 0]
	scan $m %x dec
	set id [array names Lookups ${prefix}*.$dec]
	if {[string match $id ""]} { set id UNKNOWN.-1 }
	Debugputs "ID: ..$id..  DEC: $dec M: ..$m.. SUB: $subl"
	append Telnet(comment-temp) " $Lookups($id)"

    }

    # DONT(fe) - MUST REPLY WONT fc
    # Don't get into a repeat match - after a few iterations, call it
    #  quits.
    proc fe {subl textName oobName} {
	variable Telnet
	upvar $oobName oob
	if {$Telnet(dont.[lindex $subl 0]) < 3} {
	    append oob fffc[lindex $subl 0]
	}
	addInfo $subl TELOPT_
	incr Telnet(dont.[lindex $subl 0])
	return 1
    }

    # Just walk to the end of a negotiation sequence and report
    #  where it is.

    proc findFFF0 {lst} {
	set iac 0
	set pos 0
	foreach l $lst {
	    incr pos

	    switch $l {
	    ff { set iac 1 }
	    f0 { if {$iac} { return [list $pos [lrange $lst 1 [expr $pos\
	      - 3]]] } }
	    default { set iac 0 }
	    }
	}
	error "NO END CONDITION"
    }

}

foreach index [lsort [array names Lookups]] {
    puts "$index $Telnet(lookup.$index)"
}


#
# A sample readData for socket use
#
proc readData {s} {
    set n [read $s]
    if {($n <= 0) && [eof $s]} {
	global done
	close $s
	set done 1
    }
    telnet::Debugputs "READ: [string length $n] characters"
    foreach {text oob} [telnet::splitOOB $n] { }
    telnet::Debugputs "Back from split: TXT: $text OOB: $oob"
    if {[string length $text] > 0} {puts $text}
    if {[string length $oob] > 0} {
	#telnet::Debugputs "INPUT [telnet::bin2text $n]\nREPLY:\
	  [telnet::bin2text $oob]"
	foreach cmt $telnet::Telnet(comment) {
	    puts "COMMENT: $cmt"
	}
	puts -nonewline $s $oob
	flush $s
    }
    puts -nonewline $text
}

proc bgerror {args} {
    global errorInfo
    puts "BGERR: $args \n--\n$errorInfo"
}

set site 127.0.0.1
if {$argc > 0} {
    set site $argv
}

set s [socket $argv 23]

fconfigure $s -translation binary -buffering none -blocking 0

telnet::initialize $s

fileevent $s readable "readData $s"

set done 0
vwait done


